"""Utilities for contrib module"""
# coding: utf-8

from __future__ import absolute_import as _abs
import os
import subprocess
from sys import platform as _platform
from multiprocessing import cpu_count
from ..util import TreeliteError, lineno, log_info


def _is_windows():
    return _platform == 'win32'


def _toolchain_exist_check(toolchain):
    if toolchain != 'msvc':
        retcode = subprocess.call(f'{toolchain} --version', shell=True,
                                  stdin=subprocess.DEVNULL, stdout=subprocess.DEVNULL,
                                  stderr=subprocess.DEVNULL)
        if retcode != 0:
            raise ValueError(f'Toolchain {toolchain} not found. Ensure that it is installed and '
                             + 'that it is a variant of GCC or Clang.')


def _shell():
    if _is_windows():
        return 'cmd.exe'
    if 'SHELL' in os.environ:
        return os.environ['SHELL']
    return '/bin/sh'  # use POSIX-compliant shell if SHELL is not set


def _libext():
    if _platform == 'darwin':
        return '.dylib'
    if _platform in ('win32', 'cygwin'):
        return '.dll'
    return '.so'


def _create_log_cmd_unix(logfile):
    return 'true > {}'.format(logfile)


def _save_retcode_cmd_unix(logfile):
    if _shell().endswith('fish'):  # special handling for fish shell
        return 'echo $status >> {}'.format(logfile)
    return 'echo $? >> {}'.format(logfile)


def _create_log_cmd_windows(logfile):
    return 'type NUL > {}'.format(logfile)


def _save_retcode_cmd_windows(logfile):
    return 'echo %errorlevel% >> {}'.format(logfile)


def _enqueue(args):
    tid = args['tid']
    queue = args['queue']
    dirpath = args['dirpath']
    init_cmd = args['init_cmd']
    create_log_cmd = args['create_log_cmd']
    save_retcode_cmd = args['save_retcode_cmd']

    proc = subprocess.Popen(_shell(), shell=True,
                            stdin=subprocess.PIPE,
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT)
    proc.stdin.write(init_cmd.encode())
    proc.stdin.write(f'cd {dirpath}\n'.encode())
    proc.stdin.write(create_log_cmd(f'retcode_cpu{tid}.txt' + '\n').encode())
    for command in queue:
        proc.stdin.write((command + '\n').encode())
        proc.stdin.write((save_retcode_cmd(f'retcode_cpu{tid}.txt') + '\n').encode())
    proc.stdin.flush()

    return proc


def _wait(proc, args):
    tid = args['tid']
    dirpath = args['dirpath']
    stdout, _ = proc.communicate()
    with open(os.path.join(dirpath, 'retcode_cpu{}.txt'.format(tid)), 'r') as f:
        retcode = [int(line) for line in f]
    return {'stdout': stdout.decode(), 'retcode': retcode}


# pylint: disable=R0914
def _create_shared_base(dirpath, recipe, nthread, verbose):
    # Fetch toolchain-specific commands
    obj_cmd = recipe['create_object_cmd']
    lib_cmd = recipe['create_library_cmd']
    create_log_cmd \
        = _create_log_cmd_windows if _is_windows() else _create_log_cmd_unix
    save_retcode_cmd \
        = _save_retcode_cmd_windows if _is_windows() else _save_retcode_cmd_unix

    # 1. Compile sources in parallel
    if verbose:
        log_info(__file__, lineno(),
                 'Compiling sources files in directory {} '.format(dirpath) + \
                 'into object files (*{})...'.format(recipe['object_ext']))
    ncore = cpu_count()
    ncpu = min(ncore, nthread) if nthread is not None else ncore
    workqueue = [{
        'tid': tid,
        'queue': [],
        'dirpath': os.path.abspath(dirpath),
        'init_cmd': recipe['initial_cmd'],
        'create_log_cmd': create_log_cmd,
        'save_retcode_cmd': save_retcode_cmd
    } for tid in range(ncpu)]
    for i, source in enumerate(recipe['sources']):
        workqueue[i % ncpu]['queue'].append(obj_cmd(source['name']))
    proc = [_enqueue(workqueue[tid]) for tid in range(ncpu)]
    result = []
    for tid in range(ncpu):
        result.append(_wait(proc[tid], workqueue[tid]))

    for tid in range(ncpu):
        if not all(x == 0 for x in result[tid]['retcode']):
            with open(os.path.join(dirpath, 'log_cpu{}.txt'.format(tid)), 'w') as f:
                f.write(result[tid]['stdout'] + '\n')
            raise TreeliteError('Error occured in worker #{}: '.format(tid) + \
                                '{}'.format(result[tid]['stdout']))

    # 2. Package objects into a dynamic shared library
    if verbose:
        log_info(__file__, lineno(),
                 'Generating dynamic shared library {}...' \
                 .format(
                     os.path.join(dirpath,
                                  recipe['target'] + recipe['library_ext'])))
    objects = [x['name'] + recipe['object_ext'] for x in recipe['sources']] \
              + recipe.get('extra', [])
    workqueue = {
        'tid': 0,
        'queue': [lib_cmd(objects, recipe['target'])],
        'dirpath': os.path.abspath(dirpath),
        'init_cmd': recipe['initial_cmd'],
        'create_log_cmd': create_log_cmd,
        'save_retcode_cmd': save_retcode_cmd
    }
    proc = _enqueue(workqueue)
    result = _wait(proc, workqueue)

    if result['retcode'][0] != 0:
        with open(os.path.join(dirpath, 'log_cpu0.txt'), 'w') as f:
            f.write(result['stdout'] + '\n')
        raise TreeliteError('Error occured while creating dynamic library: ' + \
                            '{}'.format(result['stdout']))

    # 3. Clean up
    for tid in range(ncpu):
        os.remove(os.path.join(dirpath, 'retcode_cpu{}.txt').format(tid))

    # Return full path of shared library
    return os.path.join(os.path.abspath(dirpath),
                        recipe['target'] + recipe['library_ext'])


__all__ = []



# coding: utf-8
"""Converter to ingest scikit-learn models into Treelite"""
import warnings

warnings.warn(('treelite.gallery.sklearn has been moved to treelite.sklearn. ' +
               'treelite.gallery.sklearn will be removed in version 1.1.'),
              FutureWarning)



# coding: utf-8
"""
Miscellaneous utilities
"""
import inspect
import ctypes
import time


class TreeliteError(Exception):
    """Error thrown by Treelite"""


def buffer_from_memory(ptr, size):
    """Make Python buffer from raw memory"""
    func = ctypes.pythonapi.PyMemoryView_FromMemory
    func.restype = ctypes.py_object
    PyBUF_READ = 0x100  # pylint: disable=C0103
    return func(ptr, size, PyBUF_READ)


def c_str(string):
    """Convert a Python string to C string"""
    return ctypes.c_char_p(string.encode('utf-8'))


def py_str(string):
    """Convert C string back to Python string"""
    return string.decode('utf-8')


@ctypes.CFUNCTYPE(None, ctypes.c_char_p)
def _log_callback(msg: bytes) -> None:
    """Redirect logs from native library into Python console"""
    print("{0:s}".format(py_str(msg)))


def lineno():
    """Returns line number"""
    return inspect.currentframe().f_back.f_lineno


def log_info(filename, linenum, msg):
    """Mimics behavior of the logging macro LOG(INFO) in dmlc-core"""
    print(f'[{time.strftime("%X")}] {filename}:{linenum}: {msg}')


def has_sklearn():
    """Check whether scikit-learn is available"""
    try:
        import sklearn  # pylint: disable=unused-import
        return True
    except ImportError:
        return False



# coding: utf-8
"""Find the path to Treelite dynamic library files."""

import os
import platform
import sys


class TreeliteLibraryNotFound(Exception):
    """Error thrown by when Treelite is not found"""


def find_lib_path():
    """Find the path to Treelite dynamic library files.

    Parameters
    ----------
    basename : :py:class:`str <python:str>`
        the base name of library
    libformat : boolean, optional (default True)
        if True, transform the base name to obtain the file name of the library
        ({}.dll on Windows; lib{}.so on Linux; lib{}.dylib on Mac OS X)
        if False, do not transform the base name at all; use it as a file name
        (this is useful to locate a file that's not a shared library)

    Returns
    -------
    lib_path: :py:class:`list <python:list>` of :py:class:`str <python:str>`
       List of all found library path to Treelite
    """
    if sys.platform == 'win32':
        lib_name = 'treelite.dll'
    elif sys.platform.startswith('linux'):
        lib_name = 'libtreelite.so'
    elif sys.platform == 'darwin':
        lib_name = 'libtreelite.dylib'
    else:
        raise RuntimeError('Unsupported operating system')

    curr_path = os.path.dirname(os.path.abspath(os.path.expanduser(__file__)))
    # List possible locations for the library file
    dll_path = [
        # normal, after installation `lib` is copied into Python package tree.
        os.path.join(curr_path, 'lib'),
        # editable installation, no copying is performed.
        os.path.join(curr_path, os.path.pardir, os.path.pardir, 'build')
    ]
    # Windows hack: additional candidate locations
    if sys.platform == 'win32':
        if platform.architecture()[0] == '64bit':
            dll_path.append(os.path.join(curr_path, '../../windows/x64/Release/'))
            # hack for pip installation when copy all parent source directory here
            dll_path.append(os.path.join(curr_path, './windows/x64/Release/'))
        else:
            dll_path.append(os.path.join(curr_path, '../../windows/Release/'))
            # hack for pip installation when copy all parent source directory here
            dll_path.append(os.path.join(curr_path, './windows/Release/'))
    # Now examine all candidate locations for the library file
    dll_path = [os.path.join(p, lib_name) for p in dll_path]
    lib_path = [p for p in dll_path if os.path.exists(p) and os.path.isfile(p)]

    if not lib_path:
        candidate_list = '\n'.join([os.path.normpath(x) for x in dll_path])
        raise TreeliteLibraryNotFound(
            f'Cannot find library {lib_name} in the candidate path: ' +
            f'List of candidates:\n{candidate_list}\n')
    return lib_path


__all__ = []



# coding: utf-8
"""Core Treelite library."""
from __future__ import absolute_import as _abs

import sys
import ctypes

import numpy as np
import scipy.sparse

from .compat import DataFrame
from .util import buffer_from_memory, c_str, py_str, _log_callback, TreeliteError
from .libpath import find_lib_path, TreeliteLibraryNotFound


def _load_lib():
    """Load Treelite Library."""
    lib_path = find_lib_path()
    lib = ctypes.cdll.LoadLibrary(lib_path[0])
    lib.TreeliteGetLastError.restype = ctypes.c_char_p
    lib.callback = _log_callback
    if lib.TreeliteRegisterLogCallback(lib.callback) != 0:
        raise TreeliteError(py_str(lib.TreeliteGetLastError()))
    return lib


# load the Treelite library globally
# (do not load if called by sphinx)
if 'sphinx' in sys.modules:
    try:
        _LIB = _load_lib()
    except TreeliteLibraryNotFound:
        _LIB = None
else:
    _LIB = _load_lib()


def _check_call(ret):
    """Check the return value of C API call

    This function will raise exception when error occurs.
    Wrap every API call with this function

    Parameters
    ----------
    ret : int
        return value from API calls
    """
    if ret != 0:
        raise TreeliteError(_LIB.TreeliteGetLastError().decode('utf-8'))


def c_array(ctype, values):
    """
    Convert a Python byte array to C array

    WARNING
    -------
    DO NOT USE THIS FUNCTION if performance is critical. Instead, use np.array(*)
    with dtype option to explicitly convert type and then use
    ndarray.ctypes.data_as(*) to expose underlying buffer as C pointer.
    """
    return (ctype * len(values))(*values)


PANDAS_DTYPE_MAPPER = {'int8': 'int', 'int16': 'int', 'int32': 'int',
                       'int64': 'int', 'uint8': 'int', 'uint16': 'int',
                       'uint32': 'int', 'uint64': 'int', 'float16': 'float',
                       'float32': 'float', 'float64': 'float', 'bool': 'i'}


def _maybe_pandas_data(data, feature_names, feature_types):
    """Extract internal data from pd.DataFrame for DMatrix data"""
    if not isinstance(data, DataFrame):
        return data, feature_names, feature_types
    data_dtypes = data.dtypes
    if not all(dtype.name in PANDAS_DTYPE_MAPPER for dtype in data_dtypes):
        bad_fields = [data.columns[i] for i, dtype in enumerate(data_dtypes) \
                      if dtype.name not in PANDAS_DTYPE_MAPPER]
        msg = "DataFrame.dtypes for data must be in, float, or bool. Did not " \
              + "expect the data types in fields "
        raise ValueError(msg + ', '.join(bad_fields))
    if feature_names is None:
        feature_names = data.columns.format()
    if feature_types is None:
        feature_types = [PANDAS_DTYPE_MAPPER[dtype.name] for dtype in data_dtypes]
    data = data.values.astype('float')
    return data, feature_names, feature_types


class DMatrix():
    """Data matrix used in Treelite.

    Parameters
    ----------
    data : :py:class:`str <python:str>` / :py:class:`numpy.ndarray` /\
           :py:class:`scipy.sparse.csr_matrix` / :py:class:`pandas.DataFrame`
        Data source. When data is :py:class:`str <python:str>` type, it indicates
        that data should be read from a file.
    data_format : :py:class:`str <python:str>`, optional
        Format of input data file. Applicable only when data is read from a
        file. If missing, the svmlight (.libsvm) format is assumed.
    missing : :py:class:`float <python:float>`, optional
        Value in the data that represents a missing entry. If set to ``None``,
        ``numpy.nan`` will be used.
    verbose : :py:class:`bool <python:bool>`, optional
        Whether to print extra messages during construction
    feature_names : :py:class:`list <python:list>`, optional
        Human-readable names for features
    feature_types : :py:class:`list <python:list>`, optional
        Types for features
    nthread : :py:class:`int <python:int>`, optional
        Number of threads
    """

    # pylint: disable=R0902,R0903,R0913

    def __init__(self, data, data_format=None, missing=None,
                 feature_names=None, feature_types=None,
                 verbose=False, nthread=None):
        if data is None:
            raise TreeliteError('\'data\' argument cannot be None')

        data, feature_names, feature_types = _maybe_pandas_data(data,
                                                                feature_names,
                                                                feature_types)
        if isinstance(data, (str,)):
            self.handle = ctypes.c_void_p()
            nthread = nthread if nthread is not None else 0
            data_format = data_format if data_format is not None else "libsvm"
            _check_call(_LIB.TreeliteDMatrixCreateFromFile(
                c_str(data),
                c_str(data_format),
                ctypes.c_int(nthread),
                ctypes.c_int(1 if verbose else 0),
                ctypes.byref(self.handle)))
        elif isinstance(data, scipy.sparse.csr_matrix):
            self._init_from_csr(data)
        elif isinstance(data, scipy.sparse.csc_matrix):
            self._init_from_csr(data.tocsr())
        elif isinstance(data, np.ndarray):
            self._init_from_npy2d(data, missing)
        else:  # any type that's convertible to CSR matrix is O.K.
            try:
                csr = scipy.sparse.csr_matrix(data)
                self._init_from_csr(csr)
            except:
                raise TypeError('Cannot initialize DMatrix from {}'
                                .format(type(data).__name__))
        self.feature_names = feature_names
        self.feature_types = feature_types
        self._get_internals()  # save handles for internal arrays

    def _init_from_csr(self, csr):
        """Initialize data from a CSR (Compressed Sparse Row) matrix"""
        if len(csr.indices) != len(csr.data):
            raise ValueError('indices and data not of same length: {} vs {}'
                             .format(len(csr.indices), len(csr.data)))
        if len(csr.indptr) != csr.shape[0] + 1:
            raise ValueError('len(indptr) must be equal to 1 + [number of rows]' \
                             + 'len(indptr) = {} vs 1 + [number of rows] = {}'
                             .format(len(csr.indptr), 1 + csr.shape[0]))
        if csr.indptr[-1] != len(csr.data):
            raise ValueError('last entry of indptr must be equal to len(data)' \
                             + 'indptr[-1] = {} vs len(data) = {}'
                             .format(csr.indptr[-1], len(csr.data)))
        self.handle = ctypes.c_void_p()
        data = np.array(csr.data, copy=False, dtype=np.float32, order='C')
        indices = np.array(csr.indices, copy=False, dtype=np.uintc, order='C')
        indptr = np.array(csr.indptr, copy=False, dtype=np.uintp, order='C')
        _check_call(_LIB.TreeliteDMatrixCreateFromCSR(
            data.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
            indices.ctypes.data_as(ctypes.POINTER(ctypes.c_uint)),
            indptr.ctypes.data_as(ctypes.POINTER(ctypes.c_size_t)),
            ctypes.c_size_t(csr.shape[0]),
            ctypes.c_size_t(csr.shape[1]),
            ctypes.byref(self.handle)))

    def _init_from_npy2d(self, mat, missing):
        """
        Initialize data from a 2-D numpy matrix.
        If ``mat`` does not have ``order='C'`` (also known as row-major) or is not
        contiguous, a temporary copy will be made.
        If ``mat`` does not have ``dtype=numpy.float32``, a temporary copy will be
        made also.
        Thus, as many as two temporary copies of data can be made. One should set
        input layout and type judiciously to conserve memory.
        """
        if len(mat.shape) != 2:
            raise ValueError('Input numpy.ndarray must be two-dimensional')
        # flatten the array by rows and ensure it is float32.
        # we try to avoid data copies if possible
        # (reshape returns a view when possible and we explicitly tell np.array to
        #  avoid copying)
        data = np.array(mat.reshape(mat.size), copy=False, dtype=np.float32)
        self.handle = ctypes.c_void_p()
        missing = missing if missing is not None else np.nan
        _check_call(_LIB.TreeliteDMatrixCreateFromMat(
            data.ctypes.data_as(ctypes.POINTER(ctypes.c_float)),
            ctypes.c_size_t(mat.shape[0]),
            ctypes.c_size_t(mat.shape[1]),
            ctypes.c_float(missing),
            ctypes.byref(self.handle)))

    def _get_dims(self):
        num_row = ctypes.c_size_t()
        num_col = ctypes.c_size_t()
        nelem = ctypes.c_size_t()
        _check_call(_LIB.TreeliteDMatrixGetDimension(self.handle,
                                                     ctypes.byref(num_row),
                                                     ctypes.byref(num_col),
                                                     ctypes.byref(nelem)))
        return (num_row.value, num_col.value, nelem.value)

    def _get_internals(self):
        data = ctypes.POINTER(ctypes.c_float)()
        col_ind = ctypes.POINTER(ctypes.c_uint32)()
        row_ptr = ctypes.POINTER(ctypes.c_size_t)()
        _check_call(_LIB.TreeliteDMatrixGetArrays(self.handle,
                                                  ctypes.byref(data),
                                                  ctypes.byref(col_ind),
                                                  ctypes.byref(row_ptr)))
        num_row, num_col, nelem = self._get_dims()

        # DMatrix should mimick scipy.sparse.csr_matrix for
        # proper duck typing in Predictor.from_csr()
        self.data = np.frombuffer(buffer_from_memory(data, ctypes.sizeof(ctypes.c_float * nelem)),
                                  dtype=np.float32)
        self.indices = np.frombuffer(buffer_from_memory(
            col_ind,
            ctypes.sizeof(ctypes.c_uint32 * nelem)), dtype=np.uint32)
        self.indptr = np.frombuffer(buffer_from_memory(
            row_ptr,
            ctypes.sizeof(ctypes.c_size_t * (num_row + 1))), dtype=np.uintp)
        self.shape = (num_row, num_col)
        self.size = nelem

    def __del__(self):
        if self.handle is not None:
            _check_call(_LIB.TreeliteDMatrixFree(self.handle))
            self.handle = None

    def __repr__(self):
        return '<{}x{} sparse matrix of type treelite.DMatrix\n' \
                   .format(self.shape[0], self.shape[1]) \
               + '        with {} stored elements in Compressed Sparse Row format>' \
                   .format(self.size)

    def __str__(self):
        # Print first and last 25 non-zero entries
        preview = ctypes.c_char_p()
        _check_call(_LIB.TreeliteDMatrixGetPreview(self.handle,
                                                   ctypes.byref(preview)))
        return py_str(preview.value)


__all__ = ['DMatrix']



# coding: utf-8
"""Converter for sklearn.ensemble.GradientBoostingClassifier (multi-class classifier)"""
import treelite


class SKLGBMMultiClassifierMixin:
    """Mixin class to implement the converter for GradientBoostingClassifier
       (multi-class classifier)"""

    @classmethod
    def process_model(cls, sklearn_model):
        """Process a GradientBoostingClassifier (multi-class classifier) to convert it into a
           Treelite model"""
        # Check for init='zero'
        if sklearn_model.init != 'zero':
            raise treelite.TreeliteError("Gradient boosted trees must be trained with "
                                         "the option init='zero'")
        # Initialize Treelite model builder
        # Set random_forest=False for gradient boosted trees
        # Set num_output_group for multi-class classification
        # Set pred_transform='softmax' to obtain probability predictions
        builder = treelite.ModelBuilder(num_feature=sklearn_model.n_features_,
                                        num_output_group=sklearn_model.n_classes_,
                                        random_forest=False,
                                        pred_transform='softmax')
        # Process [number of iterations] * [number of classes] trees
        for i in range(sklearn_model.n_estimators):
            for k in range(sklearn_model.n_classes_):
                builder.append(cls.process_tree(sklearn_model.estimators_[i][k].tree_,
                                                sklearn_model))

        return builder.commit()

    @classmethod
    def process_leaf_node(cls, treelite_tree, sklearn_tree, node_id, sklearn_model):
        """Process a test node with a given node ID"""
        leaf_value = sklearn_tree.value[node_id].squeeze()
        # Need to shrink each leaf output by the learning rate
        leaf_value *= sklearn_model.learning_rate
        # Initialize the leaf node with given node ID
        treelite_tree[node_id].set_leaf_node(leaf_value)



# coding: utf-8
"""Converter for sklearn.ensemble.RandomForestRegressor"""
import treelite


class SKLRFRegressorMixin:
    """Mixin class to implement the converter for RandomForestRegressor"""

    @classmethod
    def process_model(cls, sklearn_model):
        """Process a RandomForestRegressor to convert it into a Treelite model"""
        # Initialize Treelite model builder
        # Set random_forest=True for random forests
        builder = treelite.ModelBuilder(num_feature=sklearn_model.n_features_,
                                        random_forest=True)

        # Iterate over individual trees
        for i in range(sklearn_model.n_estimators):
            # Process the i-th tree and add to the builder
            # process_tree() to be defined later
            builder.append(cls.process_tree(sklearn_model.estimators_[i].tree_,
                                            sklearn_model))

        return builder.commit()

    @classmethod
    def process_leaf_node(cls, treelite_tree, sklearn_tree, node_id, sklearn_model):
        # pylint: disable=W0613
        """Process a test node with a given node ID"""
        # The `value` attribute stores the output for every leaf node.
        leaf_value = sklearn_tree.value[node_id].squeeze()
        # Initialize the leaf node with given node ID
        treelite_tree[node_id].set_leaf_node(leaf_value)



# coding: utf-8
"""Converter to ingest scikit-learn models into Treelite"""

from ..util import TreeliteError
from .common import SKLConverterBase
from .gbm_regressor import SKLGBMRegressorMixin
from .gbm_classifier import SKLGBMClassifierMixin
from .gbm_multi_classifier import SKLGBMMultiClassifierMixin
from .rf_regressor import SKLRFRegressorMixin
from .rf_classifier import SKLRFClassifierMixin
from .rf_multi_classifier import SKLRFMultiClassifierMixin


def import_model(sklearn_model):
    """
    Load a tree ensemble model from a scikit-learn model object

    Parameters
    ----------
    sklearn_model : object of type \
                    :py:class:`~sklearn.ensemble.RandomForestRegressor` / \
                    :py:class:`~sklearn.ensemble.RandomForestClassifier` / \
                    :py:class:`~sklearn.ensemble.GradientBoostingRegressor` / \
                    :py:class:`~sklearn.ensemble.GradientBoostingClassifier`
        Python handle to scikit-learn model

    Returns
    -------
    model : :py:class:`~treelite.Model` object
        loaded model

    Example
    -------

    .. code-block:: python
      :emphasize-lines: 8

      import sklearn.datasets
      import sklearn.ensemble
      X, y = sklearn.datasets.load_boston(return_X_y=True)
      clf = sklearn.ensemble.RandomForestRegressor(n_estimators=10)
      clf.fit(X, y)

      import treelite.sklearn
      model = treelite.sklearn.import_model(clf)
    """
    class_name = sklearn_model.__class__.__name__
    module_name = sklearn_model.__module__.split('.')[0]

    if module_name != 'sklearn':
        raise Exception('Not a scikit-learn model')

    if class_name == 'RandomForestRegressor':
        return SKLRFRegressorConverter.process_model(sklearn_model)
    if class_name == 'RandomForestClassifier':
        if sklearn_model.n_classes_ == 2:
            return SKLRFClassifierConverter.process_model(sklearn_model)
        if sklearn_model.n_classes_ > 2:
            return SKLRFMultiClassifierConverter.process_model(sklearn_model)
        raise TreeliteError('n_classes_ must be at least 2')
    if class_name == 'GradientBoostingRegressor':
        return SKLGBMRegressorConverter.process_model(sklearn_model)
    if class_name == 'GradientBoostingClassifier':
        if sklearn_model.n_classes_ == 2:
            return SKLGBMClassifierConverter.process_model(sklearn_model)
        if sklearn_model.n_classes_ > 2:
            return SKLGBMMultiClassifierConverter.process_model(sklearn_model)
        raise TreeliteError('n_classes_ must be at least 2')
    raise TreeliteError('Unsupported model type: only '
                        'random forests and gradient boosted trees are supported')


class SKLGBMRegressorConverter(SKLGBMRegressorMixin, SKLConverterBase):  # pylint: disable=C0111
    pass


class SKLGBMClassifierConverter(SKLGBMClassifierMixin, SKLConverterBase):  # pylint: disable=C0111
    pass


class SKLGBMMultiClassifierConverter(SKLGBMMultiClassifierMixin, SKLConverterBase):
    # pylint: disable=C0111
    pass


class SKLRFRegressorConverter(SKLRFRegressorMixin, SKLConverterBase):  # pylint: disable=C0111
    pass


class SKLRFClassifierConverter(SKLRFClassifierMixin, SKLConverterBase):  # pylint: disable=C0111
    pass


class SKLRFMultiClassifierConverter(SKLRFMultiClassifierMixin, SKLConverterBase):
    # pylint: disable=C0111
    pass


__all__ = ['import_model']



# coding: utf-8
"""
Treelite: a model compiler for decision tree ensembles
"""
import os

from .core import DMatrix
from .frontend import Model, ModelBuilder
from .annotator import Annotator
from .contrib import create_shared, generate_makefile, generate_cmakelists
from .util import TreeliteError
from . import sklearn

VERSION_FILE = os.path.join(os.path.dirname(__file__), 'VERSION')
with open(VERSION_FILE, 'r') as _f:
    __version__ = _f.read().strip()

__all__ = ['DMatrix', 'Model', 'ModelBuilder', 'Annotator', 'create_shared', 'generate_makefile',
           'generate_cmakelists', 'sklearn', 'TreeliteError']



# coding: utf-8
"""Converter to ingest scikit-learn models into Treelite"""
import warnings

from ...sklearn import import_model as _import_model

warnings.warn(('treelite.gallery.sklearn has been moved to treelite.sklearn. ' +
               'treelite.gallery.sklearn will be removed in version 1.1.'),
              FutureWarning)


def import_model(sklearn_model):  # pylint: disable=C0111
    return _import_model(sklearn_model)



# coding: utf-8
"""Setup Treelite package."""
import os
import shutil
import subprocess
import collections
import logging
import distutils
from platform import system
from setuptools import setup, find_packages, Extension
from setuptools.command import build_ext, sdist, install_lib, install

CURRENT_DIR = os.path.abspath(os.path.dirname(__file__))

UserOption = collections.namedtuple('UserOption', 'description is_boolean value')

USER_OPTIONS = {
    'cmake-build-dir': UserOption(description='Build directory used for CMake build',
                                  value='build', is_boolean=False)
}

NEED_CLEAN_TREE = set()
NEED_CLEAN_FILE = set()
BUILD_TEMP_DIR = None


def lib_name():
    """Return platform dependent shared object name."""
    if system() == 'Linux' or system().upper().endswith('BSD'):
        name = 'libtreelite.so'
    elif system() == 'Darwin':
        name = 'libtreelite.dylib'
    elif system() == 'Windows':
        name = 'treelite.dll'
    else:
        raise RuntimeError('Unsupported operating system')
    return name


def copy_tree(src_dir, target_dir):
    """Copy source tree into build directory."""
    logger = logging.getLogger('Treelite copy_tree')
    def clean_copy_tree(src, dst):
        logger.info('Copy tree %s -> %s', src, dst)
        distutils.dir_util.copy_tree(src, dst)
        NEED_CLEAN_TREE.add(os.path.abspath(dst))

    def clean_copy_file(src, dst):
        logger.info('Copy tree %s -> %s', src, dst)
        distutils.file_util.copy_file(src, dst)
        NEED_CLEAN_FILE.add(os.path.abspath(dst))

    cmake = os.path.join(src_dir, 'cmake')
    inc = os.path.join(src_dir, 'include')
    src = os.path.join(src_dir, 'src')

    clean_copy_tree(cmake, os.path.join(target_dir, 'cmake'))
    clean_copy_tree(inc, os.path.join(target_dir, 'include'))
    clean_copy_tree(src, os.path.join(target_dir, 'src'))

    cmake_list = os.path.join(src_dir, 'CMakeLists.txt')
    clean_copy_file(cmake_list, os.path.join(target_dir, 'CMakeLists.txt'))
    lic = os.path.join(src_dir, 'LICENSE')
    clean_copy_file(lic, os.path.join(target_dir, 'LICENSE'))


def clean_up():
    """Removed copied files."""
    for path in NEED_CLEAN_TREE:
        shutil.rmtree(path)
    for path in NEED_CLEAN_FILE:
        os.remove(path)


class CMakeExtension(Extension):  # pylint: disable=too-few-public-methods
    """Wrapper for libraries built with CMake"""
    def __init__(self, name):
        super().__init__(name=name, sources=[])


class SDist(sdist.sdist):       # pylint: disable=too-many-ancestors
    """Pack C++ sources into the Python package"""
    logger = logging.getLogger('Treelite sdist')

    def run(self):
        copy_tree(os.path.join(CURRENT_DIR, os.path.pardir),
                  os.path.join(CURRENT_DIR, 'treelite'))
        super().run()


class BuildExt(build_ext.build_ext):  # pylint: disable=too-many-ancestors
    """Custom build_ext command using CMake."""

    logger = logging.getLogger('Treelite build_ext')

    # pylint: disable=too-many-arguments,no-self-use
    def build(self, src_dir, build_dir, generator, build_tool=None):
        """Build the core library with CMake."""
        cmake_cmd = ['cmake', src_dir, generator]
        cmake_cmd.append('-DENABLE_PROTOBUF=ON')

        self.logger.info('Run CMake command: %s', str(cmake_cmd))
        subprocess.check_call(cmake_cmd, cwd=build_dir)

        if system() != 'Windows':
            nproc = os.cpu_count()
            build_cmd = [build_tool, 'treelite', '-j' + str(nproc)]
            subprocess.check_call(build_cmd, cwd=build_dir)
        else:
            subprocess.check_call(['cmake', '--build', '.', '--config', 'Release',
                                   '--target', 'treelite'], cwd=build_dir)

    def build_cmake_extension(self):
        """Configure and build using CMake"""
        src_dir = 'treelite'
        try:
            copy_tree(os.path.join(CURRENT_DIR, os.path.pardir),
                      os.path.join(self.build_temp, src_dir))
        except Exception:  # pylint: disable=broad-except
            copy_tree(src_dir, os.path.join(self.build_temp, src_dir))
        build_dir = self.build_temp
        global BUILD_TEMP_DIR  # pylint: disable=global-statement
        BUILD_TEMP_DIR = build_dir
        libtreelite = os.path.abspath(
            os.path.join(CURRENT_DIR, os.path.pardir, USER_OPTIONS['cmake-build-dir'].value,
                         lib_name()))

        if os.path.exists(libtreelite):
            self.logger.info('Found shared library, skipping build.')
            return

        self.logger.info('Building from source. %s', lib_name())
        if not os.path.exists(build_dir):
            os.mkdir(build_dir)
        if shutil.which('ninja'):
            build_tool = 'ninja'
        else:
            build_tool = 'make'

        if system() == 'Windows':
            # Pick up from LGB, just test every possible tool chain.
            for vs in ('-GVisual Studio 16 2019', '-GVisual Studio 15 2017',
                       '-GVisual Studio 14 2015', '-GMinGW Makefiles'):
                try:
                    self.build(src_dir, build_dir, vs)
                    self.logger.info(
                        '%s is used for building Windows distribution.', vs)
                    break
                except subprocess.CalledProcessError:
                    continue
        else:
            gen = '-GNinja' if build_tool == 'ninja' else '-GUnix Makefiles'
            self.build(src_dir, build_dir, gen, build_tool)

    def build_extension(self, ext):
        """Override the method for dispatching."""
        if isinstance(ext, CMakeExtension):
            self.build_cmake_extension()
        else:
            super().build_extension(ext)

    def copy_extensions_to_source(self):
        """Dummy override.  Invoked during editable installation."""
        if not os.path.exists(
                os.path.join(CURRENT_DIR, os.path.pardir, USER_OPTIONS['cmake-build-dir'].value,
                             lib_name())):
            raise ValueError('For using editable installation, please ' +
                             'build the shared object first with CMake.')


class InstallLib(install_lib.install_lib):
    """Install native library into Python package"""
    logger = logging.getLogger('Treelite install_lib')

    def install(self):

        outfiles = super().install()

        global BUILD_TEMP_DIR   # pylint: disable=global-statement

        # Copy shared library
        libtreelite_name = lib_name()
        dst_dir = os.path.join(self.install_dir, 'treelite', 'lib')
        if not os.path.exists(dst_dir):
            os.mkdir(dst_dir)
        dst = os.path.join(dst_dir, libtreelite_name)
        # CMake build dir is specified relative to the project root directory
        src_dir = os.path.join(CURRENT_DIR, os.path.pardir, USER_OPTIONS['cmake-build-dir'].value)
        if os.path.isdir(src_dir):
            # The library was built by CMake
            src = os.path.join(src_dir, libtreelite_name)
            if not os.path.exists(src):
                raise Exception(
                    f'Did not find {libtreelite_name} from directory {os.path.normpath(src_dir)}.' +
                    f' Run CMake first to build shared lib {libtreelite_name}.'
                )
            self.logger.info('Using %s built by CMake', libtreelite_name)
        else:
            # The library was built by setup.py
            build_dir = BUILD_TEMP_DIR
            src = os.path.join(build_dir, libtreelite_name)
            assert os.path.exists(src)
            self.logger.info('Using %s built by setup.py', libtreelite_name)
        self.logger.info('Installing shared library: %s -> %s', src, dst)
        dst, _ = self.copy_file(src, dst)
        outfiles.append(dst)

        # Copy VERSION
        dst_dir = os.path.join(self.install_dir, 'treelite')
        assert os.path.isdir(dst_dir)
        dst = os.path.join(dst_dir, 'VERSION')
        src = os.path.join(CURRENT_DIR, 'treelite', 'VERSION')
        assert os.path.exists(src)
        self.logger.info('Installing VERSION: %s -> %s', src, dst)
        dst, _ = self.copy_file(src, dst)
        outfiles.append(dst)

        return outfiles


class Install(install.install):  # pylint: disable=too-many-instance-attributes
    """Parse user options"""
    logger = logging.getLogger('Treelite install')
    user_options = install.install.user_options + list(
        (k + ('' if v.is_boolean else '='), None, v.description) for k, v in USER_OPTIONS.items())

    def initialize_options(self):
        super().initialize_options()
        for k, v in USER_OPTIONS.items():
            arg = k.replace('-', '_')
            setattr(self, arg, v.value)

    def run(self):
        for k, _ in USER_OPTIONS.items():
            arg = k.replace('-', '_')
            if hasattr(self, arg):
                USER_OPTIONS[k] = USER_OPTIONS[k]._replace(value=getattr(self, arg))
        super().run()


if __name__ == '__main__':
    # Supported commands:
    # From PyPI:
    # - pip install treelite
    # From source tree `treelite/python`:
    # - python setup.py install
    # - python setup.py bdist_wheel && pip install <wheel-name>
    logging.basicConfig(level=logging.INFO)
    setup(name='treelite',
          version=open(os.path.join(CURRENT_DIR, 'treelite/VERSION')).read().strip(),
          description='Treelite: model compiler for decision trees',
          install_requires=['numpy', 'scipy'],
          ext_modules=[CMakeExtension('libtreelite')],
          cmdclass={
              'build_ext': BuildExt,
              'sdist': SDist,
              'install_lib': InstallLib,
              'install': Install
          },
          extras_require={
              'scikit-learn': ['scikit-learn']
          },
          author='DMLC',
          maintainer='Hyunsu Cho',
          maintainer_email='chohyu01@cs.washington.edu',
          zip_safe=False,
          packages=find_packages(),
          include_package_data=True,
          license='Apache-2.0',
          classifiers=['License :: OSI Approved :: Apache Software License',
                       'Development Status :: 5 - Production/Stable',
                       'Operating System :: OS Independent',
                       'Programming Language :: Python',
                       'Programming Language :: Python :: 3',
                       'Programming Language :: Python :: 3.6',
                       'Programming Language :: Python :: 3.7',
                       'Programming Language :: Python :: 3.8'],
          python_requires='>=3.6',
          url='https://github.com/dmlc/treelite')

    clean_up()
